;;File Version 0.02
;;Last Modified: March 8, 2011
;;By: JEC
;;
;; >> Abstract <<
;;
;;      This file contains the conversion routines that are used in the 
;;      CASPER system. These routines allow for conversion between the
;;      I,J coordinate system, the U,V coordinate system, and the Ra and
;;      Dec system, and the Lat, Long and Radius system used in CASPER. 
;;
;; >> Details <<
;;
;;      This CASPER system uses several different coordinates systems.
;;      The following is a brief overview of each.
;;
;;       I,J system             - The data coordinate system of the plot
;;                                surface.
;;
;;       U,V system             - A scaled inertial system that is dependent
;;                                on the distance from the target to the 
;;                                observer.
;;
;;      Ra/Dec system           - Inertial spherical coordinates: Right 
;;				  Ascension and Declination.
;;
;;	Lat/Long/Radius		- Oblate spheroidal body coordinates:
;;				  Latitude, Longitude, and Radius
;;
;;	CASPER uses three different longitude systems: SPICE longitudes,
;;	"West" longitudes, and ring longitudes (referenced to an inertial
;;	direction).
;;
;; >> Contents <<
;;
;;      This file contains the following routines
;;
;;      UV_2_RD                 - Converts from UV to Ra/Dec
;;      RD_2_UV                 - Converts from Ra/Dec to UV coordinates
;;	WEST_LONG		- Converts from SPICE longitude to 
;;				  West longitude.
;;	LON_2_RLON		- Converts from body longitude to ring
;;				  longitude.
;;	RLON_2_LON		- Converts from ring longitude to target
;;				  body East longitude.
;;
;;	The following routines were added by Lynn Davis, 25 July 1996
;;
;; 	UV_2_Lat_Long		- Converts from UV to Lat/Long/Radius
;;	Lat_Long_2_UV		- Converts from Lat/Long to UV
;;	Long_Rad_2_UV		- Converts from Long/Rad to UV
;;
;;============================================================================
;;+
;;$Procedure UV_2_RD

        PRO UV_2_RD, U, V, RA, DEC, CAM_MAT 

;; >> Abstract <<
;;
;;      This subroutine is used to convert U,V coordinates into Ra and 
;;      Dec values. This involves rotating the U,V vector into an 
;;      inertial vector and then using the SPICE routine RECRAD to 
;;      get the Ra and Dec values. The values of Ra and Dec that are
;;      computed are in DEGREES. 
;;
;; >> Side Effects <<
;;
;;      The values of Ra/Dec are placed into the variables Ra, Dec
;;
;; >> Restrictions <<
;;
;;      None.
;;
;; >> Input <<
;;
;;      U       - U coordinates to be converted
;;      V       - V coordinates to be converted
;;
;; >> Output <<
;;
;;      RA      - Resulting Ra values
;;      DEC     - Resulting Dec values
;;
;;-

        UV_VEC          = dblarr(3)
;;      Begin by putting the U,V coordinates into a vector

        UV_VEC[0] = U
        UV_VEC[1] = V
        UV_VEC[2] = 1D0

;;      Rotate this vector into inertial coordinates

        CSPICE_MXV, TRANSPOSE(cam_mat), uv_vec, uvout
;;      Convert this vector into Ra and Dec

        CSPICE_RECRAD, UVout, RNG, RA, DEC

;;      Convert Ra and Dec to degress

        RA = RA * !RADEG
        DEC = DEC * !RADEG

;;      That's it, return to calling routine

        RETURN
        END

;;===========================================================================
;;+
;; Procedure RD_2_UV

        PRO RD_2_UV, RA, DEC, U, V , BEHIND , CAM_MAT

;; >> Abstract <<
;;
;;      This subroutine is used to convert Ra and Dec into U,V coordinates.
;;      Besides doing this conversion, the routine also returns a 
;;      logical flag that indicates if the position is behind the 
;;      observer. This subroutine assumes that Ra and Dec are in DEGREES.
;;
;; >> Side Effects <<
;;
;;      The values of U,V are placed into U,V. Also if the point is behind
;;      the observer, the behind flag is set to a true.
;;
;; >> Restrictions <<
;;
;;      None.
;;
;; >> Input <<
;;
;;      RA      - Ra value to be converted
;;      DEC     - Dec values to be converted
;;
;; >> Output <<
;;
;;      U       - Resulting coordinates
;;      V       - Resulting coordinates
;;      BEHIND  - Logical flag if point is behind observer
;;
;; >> Keywords <<
;;
;;      None.
;;
;;-

;;      Get the size of the passed in array
        SIZE = (n_elements(RA) > n_elements(DEC))
        uvout = DBLARR(3, size)

        FOR i = 0, size-1 DO BEGIN
            CSPICE_RADREC, 1D0, ra[i]*!DTOR, dec[i]*!DTOR, rec
            CSPICE_MXV, cam_mat, rec, uvtemp
            uvout[*,i] = uvtemp
        ENDFOR
;;      Now get the U and V coordinates 

        U = uvout[0,*] / uvout[2,*]
        V = uvout[1,*] / uvout[2,*]

;;      If UV_VEC(2) is less than 0, the position is behind the camera. 
;;      If the value is less than 0, set behind to 0.

        BEHIND = MAX(uvout[2,*] LT 0D0)

;;      That's all, return to the calling routine
        RETURN
        END

;;****************************************************************************

PRO UV_2_lat_long, U, V, Lat, Lon, Radius, RingFlag, Qrot, V_t2o

;;	U  =  the U coordinate of the point to be converted 
;;
;;	V  =  the V coordinate of the point to be converted
;;
;;	Lat = the variable to hold the returned latitude,
;;		this must be defined as a double precision
;;		floating point variable before the procedure call
;;		***this value is returned in degrees
;;
;;	Lon = the variable to hold the returned longitude,
;;		this must be defined as a double precision
;;		floating point variable before the procedure call
;;		***this value is returned in degrees
;;
;;	Radius = the variable to hold the returned radius
;;
;;	RingFlag = the variable to hold the flag that indicates
;;		that a point is on the rings of the body, it is
;;		defined as a byte integer before the procedure call.
;;


;;define some local variables

look_vec = dblarr(3)
surf_vec = dblarr(3)
nxpt = 0L             ;;"ray/plane flag" from casper_illum_ang.pro
found = 0B            ;;point can be converted is vlaue of 1
ringFlag = 0B         ;;value of 1 if on the rings
radius = 0d0
lat = 0d0
lon = 0d0
;;Create the look vector

look_vec = [U, V, 1D0]

;;Rotate the look vector into body coordinates

look_vec = Transpose(QRot) # look_vec

;;Check if the point is on the rings

;ring = On_Ring(U,V,/WRAP)

;;Point is on the ring get the point vector in body coordinates

IF (ring(0)) THEN BEGIN
   InRyPl, V_T2o, look_vec, Ring_Plane, nxpt, surf_vec
   ringFlag = 1B

;;Check if point is valid

   IF (NXPT NE 1) THEN found = 0B

;;Point is on the rings, convert to lat/lon/radius

   Reclat, surf_vec, radius, lon, lat

;;convert lat and long to degrees and make the long go from 0 to 360

   lat = lat*!RaDeg
   lon = West_Long(lon*!RaDeg)
   RETURN

ENDIF ELSE BEGIN

;;Check the point is on the body (SPICE routine)
    CASPER_GET_PLANET_CONSTANTS
    SurfPt, V_T2o, look_vec, Radii(0), Radii(1), Radii(2), surf_vec, found 

ENDELSE

;;Point isn't valid return

IF (found EQ 0) THEN BEGIN
   lat = 1000
   lon = 1000
   radius = 0
   RETURN
ENDIF

;;Point is on the body, convert to lat/long/radius

   Reclat, surf_vec, radius, lon, lat

;;convert lat and long to degrees and make the long go from 0 to 360

lat = lat*!RaDeg
lon = West_Long(lon*!RaDeg)

RETURN

END

;*****************************************************************************

PRO Lat_Long_2_UV, Lat, Lon, U, V, QROT, Target, DPV

;;*finds latitude and longitude for a point on an oblate spheroid
;;
;;	Lat =  the latitude of the point to be coverted in degrees
;;
;;	Lon = the longitude of the point to be converted in degrees
;;
;;	U  =  the variable to hold the U coordinate of the returned
;;		point, this must be defined as a double precision 
;;		floating ponit variable prior to the procedure call
;;
;;	V  =  the variable to hold the V coordinate of the returned
;;		point, this must be defined as a double precision 
;;		floating ponit variable prior to the procedure call
;;
;; >> CASPER History <<
;;
;;	  Version 2.00-T    16 May 1997     Tim Patterson
;;
;;             Updated for new, time-independent conversion routines
;;
;; 	Version 2.3	25 August 1997	Joshua Colwell
;;		Changed variable name long to lon to avoid conflict with IDL
;;		function LONG().
;;
;; Version 2.31		30 September 1997	Joshua Colwell
;;	Checks that Target Tool is open to prevent crashes with SkyMap.
;; Version 3.7		5 April 2000	Joshua Colwell
;;	remove unused common block.

;;define the latLongPt

latLongPt = dblarr(3)

;;convert the values of lat and long into a vector of 2 angles and a 
;;radius

SrfRec, target, 2.0*!dPi-Lon*!dTOr, Lat*!dTOr, latLongPt  ;SPICE

;;rotate a point in lat/long coordinates into camera coordinates

latLongPt = QRot # latLongPt

;;Convert into UV coordinates

U = (latLongPt(0)+DPV(0))/(latLongPt(2)+DPV(2))
V = (latLongPt(1)+DPV(1))/(latLongPt(2)+DPV(2))

RETURN

END

;*****************************************************************************

PRO LAT_LON_ALT_2_UV, Lat, Lon, alt, U, V, QROT, Target, DPV

;;	Lat =  the latitude of the point to be coverted in degrees
;;
;;	Lon = the longitude of the point to be converted in degrees
;;
;;	Alt = altitude above the NAIF surface of the target body, in km.
;;
;;	U  =  the variable to hold the U coordinate of the returned
;;		point, this must be defined as a double precision 
;;		floating ponit variable prior to the procedure call
;;
;;	V  =  the variable to hold the V coordinate of the returned
;;		point, this must be defined as a double precision 
;;		floating ponit variable prior to the procedure call
;;
;; >> CASPER History <<
;;
;; Version 3.7	7 April 2000	Joshua Colwell
;;	Original CASPER Version.

rectan = DBLARR(3)	;;Will hold the rectangular vector to the point in 
			;;body coordinates.

;; Convert the altitude value to a radius:
latlongpt = DBLARR(3)	;;rectangular coordinates of point on NAIF surface
			;;of body at given lat/lon.
			
SRFREC, target, 2.0*!dPi-Lon*!dTOr, Lat*!dTOr, latLongPt  ;SPICE
surface_radius = SQRT(TOTAL(latlongpt*latlongpt))
radius = surface_radius+alt

;;Now get the vector to the desired lat/lon/alt:
LATREC, radius, 2.0*!DPI-lon*!DTOR, lat*!DTOR, rectan
;;rotate a point in lat/long coordinates into camera coordinates

rectan = QRot # rectan

;;Convert into UV coordinates

U = (rectan(0)+DPV(0))/(rectan(2)+DPV(2))
V = (rectan(1)+DPV(1))/(rectan(2)+DPV(2))

RETURN

END

;*****************************************************************************

PRO Long_Rad_2_UV, Lon, Rad, U, V, qrot, dpv

;;*find the longitude and radius of a point on the rings
;;
;;	Lon = the longitude of the point to be converted in degrees
;;
;;	Rad = the radius of the point to be converted in kilometers
;;
;;	U  =  the variable to hold the U coordinate of the returned
;;		point, this must be defined as a double precision 
;;		floating ponit variable prior to the procedure call
;;
;;	V  =  the variable to hold the V coordinate of the returned
;;		point, this must be defined as a double precision 
;;		floating ponit variable prior to the procedure call
;; 
;; 	Version 2.3	25 August 1997	Joshua Colwell
;;		Changed variable name long to lon to avoid conflict with IDL
;;		function LONG().  Updated for time-dependent variables.
;;
;; Version 2.31		30 September 1997	Joshua Colwell
;;	Checks that Target Tool is open to prevent crashes with SkyMap.
;;
;;include the casper common block


;;define the latLongPt

lat = 0D0
latLongPt = dblarr(3)

lon = 2.0*!dPi-lon*!DtoR

;;convert the values of lat and long into a vector of 2 angles and a 
;;radius

LatRec, rad, lon, lat, latLongPt 	;SPICE

;;rotate a point in lat/long coordinates into camera coordinates

latLongPt = QRot # latLongPt

;;Convert into UV coordinates

U = (latLongPt(0)+DPV(0))/(latLongPt(2)+DPV(2))
V = (latLongPt(1)+DPV(1))/(latLongPt(2)+DPV(2))

RETURN

END


;;============================================================================
;;+
;;$Function LON_2_RLON

FUNCTION LON_2_RLON, lon, WEST=WEST, TIME = TIME

;;  >> Abstract <<
;;
;;      This function is used to convert target body longitude
;;	into a ring longitude.  Ring longitudes are
;;	referenced to the ascending node of the rings on Earth's equatorial
;;	plane.
;;
;;  >> Inputs <<
;;
;;	longitude to be converted, in degrees.
;;
;;  >> Keywords <<
;;
;;	WEST - when set, the input longitude is a west longitude.
;;
;;  >> Outputs <<
;;
;;	Ring longitude of the input point, in degrees.
;;
;;  >> CASPER HISTORY <<
;;
;;      Version 2.7     24 June 1998	Joshua Colwell
;;		Original Version.
;;		Funding: Cassini Science Office
;;
;;	Version 3.7	5 April 2000	Joshua Colwell
;;		Added time keyword.
;; >> Author <<
;;
;;      Joshua Colwell     - Laboratory for Atmospheric and Space Physics
;;
;;------------------------------------------------------------------------ 
;;-

pole_ra = 0D
ranums = DBLARR(3)
denums = DBLARR(3)
pole_dec = 0D
items = 0L
pole_xyz=DBLARR(3)
tipm = DBLARR(3,3)
temp = 0D
lon_ref = 0D
lat = 0D

inlon = lon

;; Get a vector pointing to the reference for ring longitude: the ascending
;; node of the ring plane on the Earth's equatorial plane.
;; First, get the pole of the target body:
IF KEYWORD_SET(TIME) THEN cent = time/100./3.16D7 ELSE $
	cent = start_et/100./3.16D7	;;Number of centuries past J2000
BODVAR, target, 'POLE_RA', items, ranums
BODVAR, target, 'POLE_DEC', items, denums
pole_ra = ranums(0)+ranums(1)*cent+ranums(2)*cent*cent
pole_dec = denums(0)+denums(1)*cent+denums(2)*cent*cent

;; Make a cartesian unit vector for the pole:
RADREC, 1D, pole_ra*!DTOR, pole_dec*!DTOR, pole_xyz

;; Make a unit vector for the pole of the Earth's equatorial plane:
earth_xyz = [0D,0D,1D]

;; Now get the vector pointing to the ring longitude reference point:
ring_ref_xyz = CROSSP(earth_xyz,pole_xyz)

IF KEYWORD_SET(WEST) THEN lon = -lon

;; Get target body longitude of the ring longitude reference direction:
IF KEYWORD_SET(TIME) THEN TIPBOD, 'J2000', target, time, tipm ELSE $
	TIPBOD, 'J2000', target, start_et, tipm
ring_ref_body_xyz = tipm#ring_ref_xyz
RECLAT, ring_ref_body_xyz, temp, lon_ref, lat

ringlon = (lon*!DTOR-lon_ref)*!RADEG

lon=inlon

RETURN,ringlon
END



;;============================================================================
;;+
;;$Function RLON_2_LON

FUNCTION RLON_2_LON, rlon, TIME = TIME

;;  >> Abstract <<
;;
;;      This function is used to convert ring longitude
;;	into a target body longitude (East).  Ring longitudes are
;;	referenced to the ascending node of the rings on Earth's equatorial
;;	plane.
;;
;;  >> Inputs <<
;;
;;	longitude to be converted, in degrees.
;;
;;  >> Outputs <<
;;
;;	East longitude of the input point, in degrees.
;;
;;  >> CASPER HISTORY <<
;;
;;      Version 2.7     24 June 1998	Joshua Colwell
;;		Original Version.
;;		Funding: Cassini Science Office
;;
;;	Version 3.7	5 April 2000	Joshua Colwell
;;		Added TIME keyword.
;; >> Author <<
;;
;;      Joshua Colwell     - Laboratory for Atmospheric and Space Physics
;;
;;------------------------------------------------------------------------ 
;;-

pole_ra = 0D
ranums = DBLARR(3)
denums = DBLARR(3)
pole_dec = 0D
items = 0L
pole_xyz=DBLARR(3)
tipm = DBLARR(3,3)
temp = 0D
lon_ref = 0D
lat = 0D

;; Get a vector pointing to the reference for ring longitude: the ascending
;; node of the ring plane on the Earth's equatorial plane.
;; First, get the pole of the target body:
IF KEYWORD_SET(TIME) THEN cent = time/100./3.16D7 ELSE $
	cent = start_et/100./3.16D7	;;Number of centuries past J2000
BODVAR, target, 'POLE_RA', items, ranums
BODVAR, target, 'POLE_DEC', items, denums
pole_ra = ranums(0)+ranums(1)*cent+ranums(2)*cent*cent
pole_dec = denums(0)+denums(1)*cent+denums(2)*cent*cent

;; Make a cartesian unit vector for the pole:
RADREC, 1D, pole_ra*!DTOR, pole_dec*!DTOR, pole_xyz

;; Make a unit vector for the pole of the Earth's equatorial plane:
earth_xyz = [0D,0D,1D]

;; Now get the vector pointing to the ring longitude reference point:
ring_ref_xyz = CROSSP(earth_xyz,pole_xyz)

;; Get target body longitude of the ring longitude reference direction:
IF KEYWORD_SET(TIME) THEN TIPBOD, 'J2000', target, time, tipm ELSE $
	TIPBOD, 'J2000', target, start_et, tipm
ring_ref_body_xyz = tipm#ring_ref_xyz
RECLAT, ring_ref_body_xyz, temp, lon_ref, lat

lon = (rlon*!DTOR+lon_ref)*!RADEG

RETURN,lon
END
